# frozen_string_literal: true

require "observer"
require "normalize_url"

module Wayfarer
  class Crawl
    extend Forwardable
    include Observable

    # The prepared job.
    # @!attribute [r] job
    attr_reader :job

    # @!attribute [r] dispatcher
    attr_reader :dispatcher

    delegate config: :job
    delegate logger: :config

    def initialize(job, *uris)
      @job = job.prepare
      @uris = uris
      @dispatcher = Dispatcher.new(@job)
      @processor = Processor.new(@job, frontier, @dispatcher)
    end

    def execute
      trap_signals

      CrawlObserver.new(@processor, @dispatcher, config.logger)

      @job.run_hook(:before_crawl)
      @processor.run(*@uris)
      @job.run_hook(:after_crawl)
    ensure
      untrap_signals
    end

    # A frontier with initially pre-staged URIs.
    # @return [Frontier]
    def frontier
      return @frontier if @frontier

      @frontier = case config.frontier
                  when :memory_trie
                    Frontiers::MemoryTrieFrontier.new(config)
                  when :redis
                    Frontiers::RedisFrontier.new(config)
                  when :memory_bloom
                    Frontiers::MemoryBloomfilter.new(config)
                  when :redis_bloom
                    Frontiers::RedisBloomfilter.new(config)
                  else
                    Frontiers::MemoryFrontier.new(config)
                  end

      @frontier.extend(Frontiers::NormalizeURIs) if config.normalize_uris

      @frontier.stage(*@uris) # TODO: Test

      @frontier
    end

    private

    def trap_signals
      @cached_sigint_handler = trap(:INT) {
        halt!
        @cached_sigint_handler.try(:call)
        exit(-1)
      }
    end

    def untrap_signals
      trap(:INT) { @cached_sigint_handler.try(:call) }
    end
  end
end
