# frozen_string_literal: true

require "redis"

module Wayfarer
  module Frontiers
    # A Redis frontier
    # @api private
    class RedisFrontier < Frontier
      def initialize(config)
        @conn = Redis.new(config.redis_opts)
        super(config)
      end

      # @override
      def current_uris
        @conn.smembers(current_uris_key).map { |str| URI(str) }
      end

      # @override
      def staged_uris
        @conn.smembers(staged_uris_key).map { |str| URI(str) }
      end

      # @override
      def stage(*uris)
        @conn.sadd(staged_uris_key, uris.map(&:to_s)) if uris.any?
      end

      # @override
      def staged?(uri)
        @conn.sismember(staged_uris_key, uri.to_s)
      end

      # @override
      def cache(*uris)
        @conn.sadd(cached_uris_key, uris.map(&:to_s)) if uris.any?
      end

      # @override
      def cached?(uri)
        @conn.sismember(cached_uris_key, uri.to_s)
      end

      # @override
      def free
        [current_uris_key, staged_uris_key, cached_uris_key].each do |key|
          @conn.del(key)
        end

        @conn.disconnect!
      end

      private

      def reset_staged_uris!
        @conn.del(staged_uris_key)
      end

      # @override
      def swap!
        # Achieve: @current_uris = @staged_uris
        @conn.rename(staged_uris_key, current_uris_key)
      end

      def filter_staged_uris!
        @conn.sdiffstore(staged_uris_key, staged_uris_key, cached_uris_key)
      end

      def current_uris_key
        "#{@config.uuid}_current_uris"
      end

      def staged_uris_key
        "#{@config.uuid}_staged_uris"
      end

      def cached_uris_key
        "#{@config.uuid}_cached_uris"
      end
    end
  end
end
