# frozen_string_literal: true

require "forwardable"
require "hooks"
require "active_job"

# TODO: I only want deep_dup
require "active_support/all"

module Wayfarer
  # A {Job} is a class that has a {Routing::Router} with many {Routing::Rule}s
  # which are matched against a URI. Rules map URIs onto job instance methods.
  # Under the hood, jobs are instantiated within separate threads by a
  # {Processor}. Every instance gets its own thread. If a URI is matched, its
  # {Page} is retrieved, and made available to instance methods via {#page}.
  #
  # Jobs implement ActiveJob's Job API and are therefore compatible with a wide
  # range of job queues. To run a job immediately, call ::perform_now.
  # enqueue a job, call ::perform_later.
  #
  # @see https://github.com/rails/rails/tree/master/activejob rails/activejob
  # @see http://edgeguides.rubyonrails.org/active_job_basics.html ActiveJob Basics
  class Job < ActiveJob::Base
    extend Forwardable

    include Hooks
    include Locals

    # @!group Callbacks

    # Callback that fires __once__ before any pages are retrieved.
    # @method before_crawl
    # @scope class
    define_hook :before_crawl

    # Callback that fires __once__ after all pages have been retrieved and
    # processing is done.
    # @method after_crawl
    # @scope class
    define_hook :after_crawl

    # Callback that fires when HTTP adapters are instantiated.
    # @method setup_adapter
    # @scope class
    # @yield [[HTTPAdapters::NetHTTPAdapter, HTTPAdapters::SeleniumAdapter], [Selenium::WebDriver::Driver, nil], [Capybara::Selenium::Driver, nil]]
    define_hooks :setup_adapter

    # @!endgroup

    class << self
      extend Forwardable

      # @!attribute [w] router
      attr_writer :router

      # @!attribute [w] config
      attr_writer :config

      # Returns a class copy.
      def prepare
        duplicate = dup
        duplicate.router = router.dup
        duplicate.locals = locals.deep_dup
        duplicate.config = config.dup

        duplicate.locals.each do |(key, val)|
          duplicate.locals[key] = Locals.thread_safe_counterpart(val)
        end

        duplicate.locals.each do |(key, _)|
          duplicate.send(:define_method, key) do duplicate.locals[key] end
          duplicate.send(:define_singleton_method, key) do
            duplicate.locals[key]
          end
        end

        duplicate
      end

      # A configuration based off the global {Wayfarer.config}.
      # @yield [Configuration]
      # @return [Configuration]
      def config
        @config ||= Wayfarer.config.clone
        yield(@config) if block_given?
        @config
      end

      # A router.
      # If a block is passed in, it is evaluated within the {Router}'s instance.
      # @return [Routing::Router]
      def router(&proc)
        @router ||= Routing::Router.new
        @router.instance_eval(&proc) if block_given?
        @router
      end

      alias route router
      alias routes router

      # Overshadows ActiveJob::Base's own logger
      delegate logger: :config
    end

    # @!attribute [r] staged_uris
    # @return [Array<String>, Array<URI>] URIs to stage for the next cycle.
    # @see #stage
    attr_reader :staged_uris

    # @!attribute [rw] page
    attr_writer :page

    # @!attribute [rw] adapter
    attr_accessor :adapter

    # @!attribute [rw] params
    attr_accessor :params

    def initialize(*argv)
      @halts = false
      @staged_uris = []
      super(*argv)
    end

    # Whether this job will stop processing.
    def halts?
      @halts
    end

    # Performs this job.
    # @note ActiveJob API
    # @override
    def perform(*uris)
      Crawl.new(self.class, *uris).execute
    end

    protected

    # All following instance methods are available within actions.

    # Sets a halting flag that signals the processor to stop its work.
    def halt
      @halts = true
    end

    # Adds URIs to process in the next cycle.
    # If a relative path is given, an absolute URI is constructed from the
    # current {#page}'s URI.
    # @param [String, URI, Array<String>, Array<URI>]
    def stage(*uris)
      expanded = uris.flatten.map do |u|
        if (uri = URI(u)).absolute?
          uri
        else
          # URI#join would discard the path of page.uri.path
          current = page.uri.dup
          current.path = File.join(page.uri.path, uri.path)
          current
        end
      end

      # This method has somewhat become the guard keeper for invalid URIs that
      # would lead to exceptions otherwise down the line
      supported = expanded.select do |uri|
        HTTPAdapters::NetHTTPAdapter::RECOGNIZED_URI_TYPES.any? do |type|
          uri.is_a?(type)
        end
      end

      @staged_uris.push(*supported)
    end

    # The {Page} representing the URI currently processed by an action.
    # When using the Selenium adapter, {Page#body} gets refreshed on every call.
    # Otherwise, subsequent DOM updates (i.e. JavaScript-induced) would be
    # invisible.
    # @return Page
    def page
      return @page unless self.class.config.http_adapter == :selenium

      Page.new(
        uri: @page.uri,
        status_code: @page.uri,
        body: driver.page_source,
        headers: @page.headers
      )
    end

    # The parsed response body.
    # When using the Selenium adapter, this parses the body again on every call.
    # Otherwise, subsequent DOM updates (i.e. JavaScript-induced) would be
    # invisible.
    # @method doc
    # @see Page#doc
    delegate doc: :page

    # The Selenium WebDriver.
    # @method driver
    # @see https://github.com/peterc/pismo Pismo
    # @see Page#driver
    delegate driver: :adapter

    # A Capybara driver that wraps the {#driver}.
    # @method browser
    # @see HTTPAdapters::SeleniumAdapter#browser
    delegate browser: :adapter

    # @method logger
    delegate logger: :"self.class"
  end
end
