# frozen_string_literal: true

require "forwardable"

module Wayfarer
  module Routing
    # Tree nodes
    # @private
    class Rule
      include Enumerable

      extend Forwardable

      attr_reader :child_rules
      attr_reader :target_action

      delegate [:each] => :child_rules

      def initialize(opts = {}, &proc)
        @child_rules = []
        @target_action = nil

        build_child_rule_chain_from_options(opts)
        instance_eval(&proc) if block_given?
      end

      def build_child_rule_chain_from_options(opts)
        @target_action = opts.delete(:to)
        opts.reduce(self) { |rule, (key, val)| rule.send(key, val) }
      end

      def matches?(uri)
        return false unless match!(uri)
        none? || any? { |child_rule| child_rule.matches?(uri) }
      end

      def invoke(uri)
        rule_chain = matching_rule_chain(uri)

        if rule_chain.any?
          params = params_from_rule_chain(rule_chain, uri)
          action = action_from_rule_chain(rule_chain)

          [true, params, action]
        else
          false
        end
      end

      # rubocop:disable Lint/AssignmentInCondition
      def matching_rule_chain(uri, chain = [])
        if match!(uri) && none?
          chain << self
        elsif matching_child = detect { |child_rule| child_rule.matches?(uri) }
          matching_child.matching_rule_chain(uri, chain << self)
        else
          []
        end
      end
      # rubocop:enable Lint/AssignmentInCondition

      def params(*)
        {}
      end

      def uri(*argv, &proc)
        append_child_rule(URIRule.new(*argv, &proc))
      end

      def host(*argv, &proc)
        append_child_rule(HostRule.new(*argv, &proc))
      end

      def path(*argv, &proc)
        append_child_rule(PathRule.new(*argv, &proc))
      end

      def query(*argv, &proc)
        append_child_rule(QueryRule.new(*argv, &proc))
      end

      def protocol(*argv, &proc)
        append_child_rule(ProtocolRule.new(*argv, &proc))
      end

      def if(*argv, &proc)
        append_child_rule(CustomRule.new(*argv, &proc))
      end

      def filetypes(*argv, &proc)
        append_child_rule(FiletypesRule.new(*argv, &proc))
      end

      private

      def append_child_rule(other)
        @child_rules << other
        other
      end

      def match!(*)
        any?
      end

      def params_from_rule_chain(rule_chain, uri)
        rule_chain.reduce({}) { |hash, rule| hash.merge(rule.params(uri)) }
      end

      def action_from_rule_chain(rule_chain)
        rule_chain.map(&:target_action).reverse.reject(&:nil?).first
      end
    end
  end
end
